function [u, out] = edgerec_pf(f, p, wTV, maxiter, etol, normtype, noise_std, u0)
% Joint image reconstruction from partial Fourier (pf) data by edge
% Original model: minimize_u wTV * |Du|_* + (1/2) * |PFu - f|^2
% Edge-based model: minimize_{uv, uh} wTV * |(uv,uh)|_* + EdgeDataFitting
% Here subscript * indicate norm type (Frobeninus, 2norm, nuclear)
%
% INPUT:
%       f: partial Fourier data of image, size = (nr, nc, nimg)
%       p: undersampling pattern in Fourier space, binary, same size as f
%       wTV: weight of edge (originally TV) term
%       maxiter: maximal number of iterations
%       etol: stopping criterion for relative changes in reconstruction
%       normtype: joint TV norm, choose from 'fro' for Frobeninus, 'mt2' for 2norm, 'nuc' for nuclear
%       noise_std: standard deviation of iid complex Gaussian noise in f
%       u0 (optional): ground truth image, size = (nr, nc, nimg)
% OUTPUT:
%       u: image
%       out: structure containing iteration number, CPU time, error, etc.
%
% prototype code on 9/17/17 by Xiaojing Ye (Georgia State Univ.)
% modified on 4/20/18 by Bin Li (South China Univ. of Tech)
% 
% copyright by Xiaojing Ye (xye@gsu.edu)
%              Bin Li (binlee@scut.edu.cn)
%
% Reference:
%   A Two-stage Algorithm for Joint Multimodal Image Reconstruction
%   Yunmei Chen, Bin Li, Xiaojing Ye
%   SIAM Journal on Imaging Sciences, 12 (3) , pp. 1425-1463, 2019.

%% initialization and pre-computation
[nr, nc, nimg] = size(f);

% initialize space for variables
uh = zeros(nr,nc,nimg); 
uv = zeros(nr,nc,nimg);

yh = zeros(nr,nc,nimg); 
yv = zeros(nr,nc,nimg);

zv = zeros(nr,nc,nimg);
zh = zeros(nr,nc,nimg);

u = zeros(nr,nc,nimg);

% initialize space for recorder
rec_cputime = zeros(1, maxiter);
rec_relerr = zeros(nimg, maxiter);

% create Fourier transform of gradient (forward finite diff) operators
FDv = psf2otf([1;-1], [nr,nc]);
FDh = psf2otf([1,-1], [nr,nc]);

% convert image Fourier data to gradient Fourier data
FDvf = zeros(nr,nc,nimg);
FDhf = zeros(nr,nc,nimg);
for k = 1:nimg
    FDvf(:,:,k) = FDv.*f(:,:,k);
    FDhf(:,:,k) = FDh.*f(:,:,k);
end

% decide if use reweighted matrix to count noise covariance
if noise_std > 0
    e_varv = max(abs(FDv).^2, 1);
    e_varh = max(abs(FDh).^2, 1);
    wTV = wTV * noise_std;
else
    e_varv = 1;
    e_varh = 1;
end

%% edge reconstruction

% initial Nesterov's acceleration parameter
t = 1; 

% main loop
for iter = 1:maxiter
    
    % set timer for current iteration
    tic;
   
    % save previous edges uv and uh
    uv_prv = uv; 
    uh_prv = uh;
    
    % prepare gradient of proximity and data terms
    for k = 1:nimg
        zv(:,:,k) = yv(:,:,k) - ifft2(p(:,:,k).*fft2(yv(:,:,k))-FDvf(:,:,k) ./ e_varv);
        zh(:,:,k) = yh(:,:,k) - ifft2(p(:,:,k).*fft2(yh(:,:,k))-FDhf(:,:,k) ./ e_varh);
    end
    zv = real(zv);
    zh = real(zh);
    
    % perform shrinkage
    switch normtype
        case 'fro'
            [uv, uh] = shrink_fro(zv, zh, wTV);
        case '2norm'
            [uv, uh] = shrink_mt2_cor(zv, zh, wTV);
        case 'nuclear'
            [uv, uh] = shrink_nuc(zv, zh, wTV);
    end
    
    % stopping criterion
    relchg_uv = abs(norm(uv(:))-norm(uv_prv(:)))/norm(uv(:));
    relchg_uh = abs(norm(uh(:))-norm(uh_prv(:)))/norm(uh(:));
    if max(relchg_uv, relchg_uh) < etol
        break; 
    end
    
    % update Nesterov's acceleration paramter t
    t_prv = t;
    t = (1 + sqrt(1+4*t^2)) / 2;
    
    % update y
    yh = uh + (t_prv-1)/t * (uh - uh_prv);
    yv = uv + (t_prv-1)/t * (uv - uv_prv);
    
    rec_cputime(iter) = toc; % cpu time used for current iteration
    
    % compare to ground truth (computation time not counted)
    if nargin == 8 % ground truth image u0 available
        u = edge2img_pf(f,p,uv,uh,1e-2);
        for k = 1:nimg
            rec_relerr(k, iter) = norm(u(:,:,k)-u0(:,:,k),'fro')/norm(u0(:,:,k),'fro');
        end
    end
    
end

u = edge2img_pf(f,p,uv,uh,1e-2);

%% other outputs

out.iter = iter;
out.cputime = cumsum(rec_cputime,2);
out.cputime = [0, out.cputime(1:iter-1)];
out.relerr = [ones(nimg,1), rec_relerr(:, 1:iter-1)];